#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""braise - a dynamic firewall

"""

__all__ = ["no_arg_provided", "enable", "disable", "show_logs", "show_realms", "show_services", "show_service", "show status", "add_service", "del_service"]


import pickle
# Needed to access logging :
from systemd import journal
import select
#import putch
import datetime

# Needed to add/del services :
from servicewall import servicewall
from servicewall import network_helpers

# Needed to parse arguments :
import argparse
#import argcomplete

import os


conf_dir = "/var/lib/servicewall/"
realm_defs_pickle = conf_dir + "realms.p"
service_defs_pickle = conf_dir + "services.p"
definitions_dir = "/etc/gufw/app_profiles"
program_name = "ServiceWall"

with open(realm_defs_pickle, "rb") as fd:
    realm_defs = pickle.load(fd)
with open(service_defs_pickle, "rb") as fd:
    service_defs = pickle.load(fd)

def helper(func, args):
    #wrapper = eval func.__name__
    #wrapper(args)
    func(args)

def print_dict(dictionary):
    for key, value in dictionary.items():
        if isinstance(value, dict):
            print("=> %s :" % key)
            print_dict(value)
        else:
            print("%s : %s" % (key, value))


def no_arg_provided(args):
    #parser.print_help()
    raise SystemExit("\n  argument needed !")

def enable(args):
    src = "/etc/NetworkManager/dispatcher.d/toggler"
    dst = "/usr/lib/servicewall/toggler"
    if os.path.exists(src):
        print("Already enabled")
    else:
        os.symlink(dst, src)
    firewall = servicewall.ServiceWall()
    if not firewall.up:
        firewall = servicewall.ServiceWall()
        firewall.start()
        print("Firewall started")

def disable(args):
    target = "/etc/NetworkManager/dispatcher.d/toggler"
    # DEBUG This test would fail on a broken link :
    if os.path.exists(target):
        os.remove(target)
    else:
        print("Already disabled")
    firewall = servicewall.ServiceWall()
    if firewall.up:
        firewall.stop()
        print("Firewall stopped")


def show_status(args):
    firewall = servicewall.ServiceWall()
    firewall.list_services_in()

def show_realms(args):
    print_dict(realm_defs)
def show_services(args):
    for service in service_defs:
        print("%s - %s" % (service, service_defs[service]["description"]))
def show_service(args):
    #service_name = args.service_name
    # The calling parser has nargs="*", so args.service_name is a list
    service_name = " ".join(args.service_name)
    # Do our own validity testing
    if service_name not in service_defs:
        #parser_show_service.print_usage()
        raise SystemExit('service "%s" not found. ')
    print_dict(service_defs[service_name])

def add_service(args):
    service_name = args.service_name
    print(service_name, str(len(service_name)))
    # Do our own validity testing
    if service_name not in service_defs:
        raise SystemExit('service "%s" not found. ')
    essid = network_helpers.get_essid()
    if service_name in realm_defs[essid]:
        raise SystemExit("Service %s already in realm %s's definition" %
                (service_name, essid))
    realm_defs[essid][service_name] = False
    with open(realm_defs_pickle, "wb") as fd:
        pickle.dump(realm_defs, fd)
    firewall = servicewall.ServiceWall()
    firewall.add_service_in(service_name, local=False)
    print("Allowed %s to be served when in realm %s." % (service_name, essid))

def del_service(args):
    service_name = args.service_name
    # Do our own validity testing
    if service_name not in service_defs:
        raise KeyError('service "%s" not found. ')
    essid = network_helpers.get_essid()
    if service_name in realm_defs[essid]:
        del realm_defs[essid][service_name]
    else:
        raise KeyError('service "%s" not allowed in realm %s anyway.'
                % (service_name, essid))
    with open(realm_defs_pickle, "wb") as fd:
        pickle.dump(realm_defs, fd)
    firewall = servicewall.ServiceWall()
    firewall.del_service_in(service_name)
    print("Removed %s from allowed services on realm %s." % (service_name, essid))

def show_logs(args):
    if "period" in args:
        yielder = log_yielder(int(args.period))
    else:
        yielder = log_yielder()
    now = datetime.datetime.today()

    # log_folder is a dict that contains host adresses as keys,
    # and whose items are dicts that contain port numbers as keys,
    # and whose items are lists of logs, which are dicts.
    #
    # log_folder {
    #             "192.168.1.1": {
    #                             "22": {
    #                                    [log1, log2, ... ] }}}
    #
    # log1: {
    #         "DATE": datetime.timestamp, "SRC": "191.168.1.1", ... }

    log_folder = {}

    for log in yielder:
        if log["SRC"] not in log_folder:
            log_folder[log["SRC"]] = {}
        if "DPT" not in log:
            log["DPT"] = "-1"
        if log["DPT"] not in log_folder[log["SRC"]]:
            log_folder[log["SRC"]][log["DPT"]] = [ log, ]
        else:
            log_folder[log["SRC"]][log["DPT"]].append(log)


    print("Hosts : %s" % ", ".join(ip for ip in log_folder.keys()))
    for src, ports in log_folder.items():
        print("Host %s" % src)
        for dpt, logs in ports.items():
            age = str(now - logs[0]["DATE"]).split(".")[0]
            if dpt == "-1":
                # These packets don't have any destination port.
                # We'll print the first, and any variations on it.
                variations = []
                for log in logs[1:]:
                    for key in log:
                        if key in ("DATE", "ID"):
                            continue
                        if log[key] != logs[0][key]:
                            variations += (key, log[key])
                # Delete "DPT" key as it's not a relevant value :
                del logs[0]["DPT"]
                print("  sent %i unknow packet(s) until %s ago :" %
                        (len(logs), age))
                del logs[0]["DATE"]
                del logs[0]["SRC"]
                for key, value in logs[0].items():
                    print("    %s: %s" % (key, value))
                if variations:
                    print("  variations :")
                    for var in variations:
                        print("    " + var)
            else:
                print("  asked %i times for port %s until %s ago" %
                    (len(logs), dpt, age))

def log_yielder(period=""):
    """get logs we implemented in iptables from journald"""
    reader = journal.Reader()
    reader.log_level(journal.LOG_WARNING)
    reader.add_match(SYSLOG_IDENTIFIER="kernel")
    now = datetime.datetime.today()
    #p = select.poll()
    #p.register(reader, reader.get_events())
    #p.poll()
    reader.seek_tail()
    while True:
        log = reader.get_previous()
        # Quit if there's no message :
        if not "MESSAGE" in log:
            break
        # Quit if log older than period :
        if period:
            age = datetime.datetime.timestamp(now) - datetime.datetime.timestamp(log["__REALTIME_TIMESTAMP"])
            if age > period:
                break
        # Only catch messages sent by iptables log :
        if log["MESSAGE"].startswith(program_name):
            message_dict = {}
            message = log["MESSAGE"].strip(program_name + ":").strip()
            message_dict["DATE"] = log["__REALTIME_TIMESTAMP"]
            for item in message.split():
                if item.count("="):
                    key, value = item.split("=")
                    message_dict[key] = value
                else:
                    message_dict[item] = ""
            yield message_dict





"""Useful system calls to ask different things to the wireless controler.
"""


import socket
import struct
import array
import fcntl
import netifaces


def get_active_interface():
    """Return the name of the interface serving as gateway.
    """
    # AF_INET is a flag for ipv4 addressing
    return netifaces.gateways()[netifaces.AF_INET][0][1]

def get_essid():
    """Return the ESSID for an interface, or None if we aren't connected.
    """
    interface = get_active_interface()
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # The payload for this call is formatted to 32 double hexadecimal ints.
    essid = array.array("b", b"\x00" * 32)
    essidPointer, essidLength = essid.buffer_info()
    request = array.array("b")
    request.frombytes(
            # The interface name is formatted to 16 double hexadecimal ints.
            interface.ljust(16, "\x00").encode()
            + struct.pack("PHH", essidPointer, essidLength, 0)
    )
    fcntl.ioctl(
            s.fileno(),
            0x8b1b,   # SIOCGIWESSID, get essid
            request)
    name = essid.tobytes().strip(b"\x00").decode()
    if name:
        return name
    return None

def get_ip_address(ifname):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack("256s", bytes(ifname[:15], "utf-8"))
    )[20:24])

def get_netmask(ifname):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x891b, # SIOCGIFNETMASK
            struct.pack("256s", bytes(ifname[:15], "utf-8"))
    )[20:24])


def get_subnetwork():
    """return the IP address of the subnetwork slash the netmask, as in

            "192.168.1.0/255.255.255.0"
    """
    interface = get_active_interface()
    address = get_ip_address(interface).split(".")
    netmask = get_netmask(interface).split(".")
    subnetwork = [ str(int(int(address_block) * int(netmask_block) / 255))
            for address_block, netmask_block in zip(address, netmask) ]
    return "/".join((".".join(subnetwork), ".".join(netmask) ))

def get_essid_mac_address(ifname):
    """Return the ssid's MAC address as declared by the network controller
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    payload = fcntl.ioctl(
            s.fileno(),
            0x8b15, # SIOCGIWAP , mac address
            struct.pack("265s", bytes(ifname[:15], "utf-8"))
    )
    # From internet talk : SSID is encoded after a \x01\x00 keycode.
    index = payload.index(b"\x01\x00") + 2
    # From convention : MAC addressing is 6 hexadecimal double ints.
    address = payload[index:index+6].hex()
    return ":".join([ address[i:i+2] for i in range(0, 12, 2) ])

def get_essid_alt(ifname):
    from subprocess import Popen, PIPE
    process = Popen(["iw", ifname, "info"], stdout=PIPE)
    stdout, stderr = process.communicate()
    try:
        endstring = stdout[stdout.index(b"ssid ") + 5:]
    except ValueError:
        raise SystemExit("Interface %s not found" % ifname)
    ssid = endstring.split(b"\n")[0]
    return ssid.decode()

def get_gateway_address():
    """Read the default gateway directly from /proc."""
    with open("/proc/net/route") as fh:
        for line in fh:
            fields = line.strip().split()
            if fields[1] != '00000000' or not int(fields[3], 16) & 2:
                continue
            return socket.inet_ntoa(struct.pack("<L", int(fields[2], 16)))

def get_gateway_hostname():
    return socket.getfqdn(get_gateway_address())






"""Firewall class

implements a firewall using python-iptables
"""

from os import environ
from iptc import Rule, Match, Chain, Table


identifier = "ServiceWall"


class FireWall():
    """A simple firewall class

    In this firewall, there is (only) one _table called FILTER. It contains
    2 chains called input and output. When anything comes from outside, it
    goes through input. When anything goes outside, it goes through output.

    You can add (and delete) rules based on any valid iptables match.

    TODO :
    - ruleset Object.
    """
    def __init__(self):
        self._table = Table(Table.FILTER)

        # _table.autocommit is True by default.
        #self._table.autocommit = False
        self.input_chain = Chain(self._table, "INPUT")
        self.output_chain = Chain(self._table, "OUTPUT")
        if self.status():
            self.up = True
        else:
            self.up = False

    def start(self):
        """Start a basic FireWall.

        Drops all incoming, accepts only localhost.
        Log anything that is dropped
        """
        # Drop all incoming. Basic but functional.
        self.input_chain.set_policy("DROP")
        self.output_chain.set_policy("ACCEPT")
        # Accept all on localhost.
        self.add_rule(
                "localhost",
                self.input_chain,
                "ACCEPT",
                iface="lo"
                )
        # Log all that is refused.
        self.add_rule(
                "log",
                self.input_chain,
                "LOG"
                )
        if not self._table.autocommit:
            self._table.commit()
        self.up = True

    def stop(self):
        """Remove rules.

        This function removes all rules created by this firewall's instances ;
        it recognizes them by the comment that's inside of them.
        """
        found_rules = 0
        for rule in self.input_chain.rules:
            name = self._get_rule_name(rule)
            ident = self._get_rule_id(rule)
            if ident == identifier:
                self.del_rule(self._get_rule_name(rule), self.input_chain)
                found_rules += 1
        if not found_rules:
            print("no rule found for id %s." % identifier)
        print("setting input policy to ACCEPT")
        self.input_chain.set_policy("ACCEPT")
        if not self._table.autocommit:
            self._table.commit()
        self.up = False

    def status(self):
        """print the status of the FireWall. Returns either True or False."""
        # Returns True if a single rule has our identifier tag
        for rule in self.input_chain.rules:
            if identifier == self._get_rule_id(rule):
                return True
        else:
            return False

    def add_rule(self, name, chain, target, dst="", dport="", src="", sport="", proto="", iface=""):
        """Add rule.

        chain should be either self.input_chain or self.output_chain.
        traget is a string saying what to do, like
                "ACCEPT", "DENY", "DROP"
        proto (if not defined, is assumed both udp and tcp)
        src is a realm in the form XXX.XXX.XXX.XXX/YY
        sport is source port. Currently ignored
        dst is a realm in the same form as src
        dport is destination port
        iface if not defined, will match any
        """
        print("adding rule %s" % name)
        if (dport or sport) and not proto:
            for proto in "tcp", "udp":
                self.add_rule(name, chain, target, dst, dport, src, sport, proto, iface)
        else:
            rule = Rule()
            rule.create_target(target)
            if target == "LOG":
                limit_match = rule.create_match("limit")
                limit_match.limit = "1/s"
                limit_match.limit_burst = "1"
                rule.target.set_parameter("log-prefix", identifier + ":")
            # First we need to know if we go in or out
            # Then some rules need the creation of a match
            if dst:
                rule.dst = dst
            if dport:
                proto_match = rule.create_match(proto)
                proto_match.dport = str(dport)
                rule.protocol = proto
            if src:
                rule.src = src
            if sport:
                proto_match = rule.create_match(proto)
                proto_match.sport = str(sport)
                rule.protocol = proto
            if iface:
                rule.in_interface = iface
                rule.out_interface = iface
            # Add a signature as a comment.
            comment_match = rule.create_match("comment")
            comment_match.comment = identifier + ":" + name
            # Try to set it into chain.
            rule.final_check()
            if target == "LOG":
                chain.append_rule(rule)
            else:
                chain.insert_rule(rule)
            if not rule in chain.rules:
                print("Need to check %s rule." % name)

    def del_rule(self, name, chain):
        """Remove rule.

        This function removes one rule named "name" in chain "chain". It
        expects the rule to contain a comment in the form
            "identifier:name"
        where identifier is the name of the program that has set the rule, and
        title is the name of the rule. Note that there may be several rules
        under the same name. This should delete the first it finds.
        """
        for rule in chain.rules:
            if self._get_rule_name(rule) == name:
                chain.delete_rule(rule)
                print("deleted rule %s" % name)
                break

    def list_rules(self, chain):
        for rule in chain.rules:
            # First find out dport - it gets set by self.add_service_in as a match.
            dport = "any"
            proto = "any"
            for match in rule.matches:
                try:
                    dport = match.parameters["dport"]
                    proto = match.name
                except (KeyError, IndexError):
                    pass
            # Then find out source.
            if rule.src == "0.0.0.0/0.0.0.0":
                src = "any"
            else:
                src = rule.src
            print("%s %s\t\t: %s port %s from %s" %
                    (rule.target.name, self._get_rule_name(rule),
                        proto, dport, src))

    def _get_rule_name(self, rule):
        for match in rule.matches:
            if match.comment:
                try:
                    rule_identifier, rule_name = match.comment.split(":")
                except IndexError:
                    # No semicolon in comment
                    continue
                return rule_name

    def _get_rule_id(self, rule):
        for match in rule.matches:
            if match.comment:
                try:
                    rule_identifier, rule_name = match.comment.split(":")
                except IndexError:
                    # No semicolon in comment
                    continue
                return rule_identifier




from os import scandir


def split_port_def(port_def):
    ports_list = []
    # In the worst case, port_def is "13:15,124,128/udp|120:122/tcp"
    # Which should mean 13 udp, 124 udp and range 120-122 tcp
    try:
        port_subdefs = port_def.split("|")
    except ValueError:
        port_subdefs = (port_def, )

    # first port_subdef is "13:15,124,128/udp"
    for port_subdef in port_subdefs:
        try:
            port_subdef, proto = port_subdef.split("/")
        except ValueError:
            # We have a port but no proto ; that's OK.
            # Empty proto means any of udp or tcp.
            proto = ""

        # first port_subdef is now "13:15,124,128" : proto is "udp"
        try:
            ports = port_subdef.split(",")
        except ValueError:
            ports = (port_subdef, )

        # ports is ("13:15", "124", "128")
        for port in ports:
            # port is a string ; let's check if this str is a number
            if port.isalnum():
                ports_list += ((int(port), proto), )
            else:
                # It's a port range, spelled ie "100:104".
                ports_start, ports_end = port.split(":")
                for port_range_item in range(int(ports_start), int(ports_end)+1):
                    ports_list += ((port_range_item, proto), )

    return tuple(ports_list)


def scan_service_definitions(definitions_dir):
    '''scan service definitions in the list given by jhansonxi'''
    # Files in the following dir define one or more protocols and
    # their portsi. Each file comes with the following format :
    #
    # [NAME_ONE]
    # title=EXPANDED NAME
    # description=BLA BLA BLA
    # categories=ONE;TWO;
    # reference=[HTTP://HTML_LINK LINK_NAME DESCRIPTION]
    # ports=220|3724,6112:6114,4000/tcp
    #
    # [NAME_TWO]
    # etc...

    filelist = scandir(definitions_dir)
    services = {}
    for filename in iter(filelist):
        print("processing %s" % filename.path)
        with open(filename.path, "r") as fd:
            # NAME_ONE and NAME_TWO are separated by a double newline
            service_list = fd.read().split("\n\n")
            for service in service_list:
                # _Some_ service descriptions finish in "\n\n"
                if service == "" or service == "\n":
                    continue
                service_desc = service.split("\n")
                service_name = service_desc[0].strip("[]")

                # get rid of empty strings :
                for i in range(service_desc.count("")):
                    service_desc.remove("")

                # services['NAME_ONE'] will return a dict with its parameters :
                services[service_name] = {
                    key: value for (key, value) in
                        ( item.split("=", 1) for item in iter(service_desc[1:]) )
                }
                port_def = split_port_def(services[service_name]["ports"])
                services[service_name]["ports"] = port_def

    return services





"""  ServiceWall

Uses service definitions provided by jhansonxi
and implements them in a FireWall class, either to allow them for the local
subnetwork, or worldwide.
"""

import pickle
from servicewall import network_helpers
from servicewall import statefulfirewall


class ServiceWall(statefulfirewall.StateFulFireWall):
    """ServiceWall - a FireWall in which you can add services on the fly.
    """

    identifier = "ServiceWall"
    conf_dir = "/var/lib/servicewall/"
    service_defs_pickle = conf_dir + "services.p"
    realm_defs_pickle = conf_dir + "realms.p"

    def __init__(self):
        super().__init__()
        # TODO if the FW is already up, we need to check what was the
        # essid it connected to, to see if we need to reload.
        self.essid = network_helpers.get_essid()
        self.subnetwork = network_helpers.get_subnetwork()
        with open(self.service_defs_pickle, "rb") as fd:
            self.service_defs = pickle.load(fd)
        with open(self.realm_defs_pickle, "rb") as fd:
            self.realm_defs = pickle.load(fd)

    def start(self, **args):
        """Will load a set of rules from self.realm_defs . If these rules
        are set as True, they are matched with the provided subnetwork,
        else to any source.
        """
        if self.essid not in self.realm_defs:
            # If we don't have a definition in there, load "FireWall:new"
            self.realm_defs[self.essid] = self.realm_defs[identifier + ":new"]
        for service_name, local_toggle in self.realm_defs[self.essid].items():
            if local_toggle:
                self.add_service_in(service_name, local=True)
            else:
                self.add_service_in(service_name, local=False)
        super().start(**args)   # commits the table if relevant

    def save(self):
        """Dumps the actual config to config file.
        """
        with open(realm_defs_pickle, "wb") as fd:
            pickle.dump(self.realm_defs, fd)
        print("Modified realm rules for %s written to file." % self.essid)

    def add_service_in(self, service_name, local=False):
        """Open ports for a service hosted on this machine.
        
        service_name should be one of self.service_defs' keys.
        if src is "local", use self.subnetwork instead.
        """
        if service_name not in self.service_defs:
            raise KeyError("service_defs knows no service service_named %s."
                    % service_name)

        if local:
            src = self.subnetwork
        else:
            src = ""

        for port, proto in self.service_defs[service_name]["ports"]:
            self.add_rule(
                    service_name,
                    self.input_chain,
                    "ACCEPT",
                    src=src,
                    dport=port,
                    proto=proto,
            )

    def del_service_in(self, service_name):
        """Closes ports for service service_name if they were opened.
        """
        for rule in self.input_chain.rules:
            # Call the FireWall's  private _get_rule_name function
            if super()._get_rule_name(rule) == service_name:
                #self.del_rule(service_name, self.input_chain)
                self.input_chain.delete_rule(rule)


    def list_services_in(self):
        """Lists services for which we have allowed ports.
        """
        self.list_rules(self.input_chain)



#!/usr/bin/env python


import setuptools
from setuptools.command.install import install
import os
from os import path, environ

class CustomInstallCommand(install):
    def run(self):
        #dst = os.readlink(here + "/links/toggler")
        src = "/etc/NetworkManager/dispatcher.d/toggler"
        srcpath = "/".join(src.split("/")[:-1])
        print("=> %s %s" % (srcpath, path.exists(srcpath)))
        #for key, item in environ.items():
        #    print(key, item)
        install.run(self)

with open("README.md", "r") as fh:
    long_description = fh.read()

#   4 - Beta
#   5 - Production/Stable

here = path.abspath(path.dirname(__file__))
print("included packages : %s" %
        setuptools.find_packages(exclude=['scriptlets']))
name="servicewall"
version = "0.3.1"

setuptools.setup(
    name=name,
    version=version,
    author="la Fleur",
    author_email="lafleur@boum.org",
    description="the desktop firewall that adapts to different network connections",
    keywords="dynamic adaptable iptables firewall",
    license="GNUv3",
    python_requires=">=3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://flip.local/~lafleur/servicewall",
    packages=setuptools.find_packages(exclude=['scriptlets']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=["python-iptables"],
    scripts=["servicewall/braise"],
    data_files=[
        ("/var/lib/servicewall", ["var/realms.p", "var/services.p"]),
        ("/usr/lib/servicewall", ["servicewall/toggler"]),
    ],
    cmdclass={"install": CustomInstallCommand,},
)

# Update PKGBUILD's md5sums
try:
    from update_md5sum_in_PKGBUILD import do_md5_sum
    do_md5_sum("dist/" + name + "-" + version + ".tar.gz")
except ModuleNotFoundError:
    pass



"""  StateFulFireWall

Uses simple guidelines given in

and implements them in a FireWall class, using reasonable defaults.
"""

from iptc import Rule
from servicewall import firewall


class StateFulFireWall(firewall.FireWall):
    """Implement some useful stateful rules :
    
    - accept related/established packets
    - drop invalid packets
    """

    identifier = "ServiceWall"

    def __init__(self):
        super().__init__()

    def start(self, **args):

        print("adding rule icmp")
        rule = Rule()
        rule.create_target("ACCEPT")
        rule.protocol = "icmp"
        m = rule.create_match("icmp")
        m.set_parameter("icmp-type", "8")
        comment_match = rule.create_match("comment")
        comment_match.comment = self.identifier + ":icmp"
        self.input_chain.append_rule(rule)

        # Top rule should be to allow related, established connections.
        self.add_conntrack_rule_in("ACCEPT", "ctstate", "RELATED,ESTABLISHED")

        # Drop invalid packets - as diagnosed by the conntrack processor.
        self.add_conntrack_rule_in("DROP", "ctstate", "INVALID", position="bottom")
        super().start(**args)   # commits the table if relevant

    def add_conntrack_rule_in(self, target, param_key, param_value, position="top"):
        """Adds a rule to input chain, for example :

        StateFulFireWall.add_conntrack_rule_in("ACCEPT", "ctstate", "RELATED")
        """
        print("adding rule %s" % param_value)
        conntrack_rule = Rule()
        conntrack_rule.create_target(target)
        conntrack_match = conntrack_rule.create_match("conntrack")
        conntrack_match.set_parameter(param_key, param_value)
        comment_match = conntrack_rule.create_match("comment")
        comment_match.comment = self.identifier + ":" + param_value
        if position == "top":
            self.input_chain.insert_rule(conntrack_rule)
        elif position == "bottom":
            self.input_chain.append_rule(conntrack_rule)


